package com.bcxin.tenant.data.etc.tasks.jobs;
import com.bcxin.event.core.exceptions.BadEventException;
import com.bcxin.flink.streaming.cores.properties.CheckpointConfigProperty;
import com.bcxin.tenant.data.etc.tasks.components.*;
import com.bcxin.tenant.data.etc.tasks.properties.DataEtcConfigProperty;
import org.apache.commons.collections.CollectionUtils;
import org.apache.flink.api.common.JobExecutionResult;
import org.apache.flink.api.common.eventtime.WatermarkStrategy;
import org.apache.flink.connector.jdbc.JdbcExecutionOptions;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.connector.kafka.source.enumerator.initializer.OffsetsInitializer;
import org.apache.flink.connector.kafka.source.reader.deserializer.KafkaRecordDeserializationSchema;
import org.apache.flink.core.execution.JobClient;
import org.apache.flink.core.execution.JobListener;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.OffsetResetStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

public class DataEtcJob extends DataJobAbstract {
    private static final Logger logger = LoggerFactory.getLogger(DataEtcJob.class);

    private final Collection<DataEtcConfigProperty> configProperties;
    private final String primaryKeyName = "id";
    private final String env;
    private final String configFile;
    private final boolean isDebug;
    private final CheckpointConfigProperty checkpointConfigProperty;

    public DataEtcJob(CheckpointConfigProperty checkpointConfigProperty,
                      Collection<DataEtcConfigProperty> configProperties,
                      String env, String configFile, boolean isDebug) {
        super(configProperties, configFile, isDebug);
        this.checkpointConfigProperty = checkpointConfigProperty;
        this.configProperties = configProperties;
        this.env = env;
        this.configFile = configFile;
        this.isDebug = isDebug;
    }

    @Override
    protected void coreExecute() throws Exception {
        super.coreExecute();

        StreamExecutionEnvironment env = getStreamExecutionEnvironment(this.checkpointConfigProperty);

        KafkaRecordDeserializationSchema<BinlogRawValue> kafkaRecordDeserializationSchema = getDeserializationSchema();

        JdbcExecutionOptions jdbcExecutionOptions =
                JdbcExecutionOptions.builder()
                        .withBatchSize(3_000)
                        .withBatchIntervalMs(60_000)
                        .withMaxRetries(3)
                        .build();

        for (DataEtcConfigProperty config : configProperties) {
            buildConfigSourceAndSink(env, kafkaRecordDeserializationSchema, jdbcExecutionOptions, config);
        }

        String title = configProperties.stream()
                .map(ix -> ix.getTitle()).distinct()
                .collect(Collectors.joining(";"));
        env.registerJobListener(new JobListener() {
            @Override
            public void onJobSubmitted(@Nullable JobClient jobClient, @Nullable Throwable throwable) {
                if (jobClient != null) {
                    logger.error("Job以及提交到服务端: jobId={};status={};", jobClient.getJobID(), jobClient.getJobStatus());
                }
            }

            @Override
            public void onJobExecuted(@Nullable JobExecutionResult jobExecutionResult, @Nullable Throwable throwable) {
                if (jobExecutionResult != null) {
                    logger.error("Job以及提交到服务端: jobId={};JobExecutionResult={};",
                            jobExecutionResult.getJobID(),
                            jobExecutionResult.isJobExecutionResult()
                    );
                }
            }
        });

        env.execute(String.format("job-v18.3:%s-环境-%s-%s", this.env, title, configFile));
    }

    @Override
    public String getRegionPrefix() {
        DataEtcConfigProperty property = this.configProperties.stream().findFirst().orElse(null);
        if(property != null){
            return property.getRegionPrefix() == null ? "" : property.getRegionPrefix();
        }
        return "";
    }

    private void buildConfigSourceAndSink(StreamExecutionEnvironment env,
                                          KafkaRecordDeserializationSchema<BinlogRawValue> kafkaRecordDeserializationSchema,
                                          JdbcExecutionOptions jdbcExecutionOptions,
                                          DataEtcConfigProperty config) {
        if (config == null || CollectionUtils.isEmpty(config.getTopicSubscribers())) {
            throw new BadEventException("无效KafkaTopic配置数据信息");
        }

        Collection<DataEtcConfigProperty.TmpMergedKafkaConnectionTopicInfo>
                kafkaConnectionTopicInfos = getMergedKafkaConnections(config);

        kafkaConnectionTopicInfos.forEach(kti -> {
            Collection<DataEtcConfigProperty.TopicSubscriberConfigProperty> selectedTopicSubscriberConfigs =
                    config.getTopicSubscribers().stream()
                            .filter(ix -> ix.getRefKafkaName().equalsIgnoreCase(kti.getKafkaConnection().getName()) &&
                                    kti.getTopics().stream().anyMatch(ii -> ii.equalsIgnoreCase(ix.getTopic()))
                            ).collect(Collectors.toList());

            if (CollectionUtils.isEmpty(selectedTopicSubscriberConfigs)) {
                throw new BadEventException("无效主题配置");
            }

            DataEtcConfigProperty.TopicSubscriberConfigProperty selectedSubscriberConfig
                    = selectedTopicSubscriberConfigs.stream().findFirst().get();
            if (selectedTopicSubscriberConfigs.size() > 1) {
                throw new BadEventException(String.format("同一个主题(%s)在一个文件中必须只能只有一个主题订阅",
                        selectedSubscriberConfig.getTopic()));
            }

            String[] selectedTopics = kti.getTopics().toArray(new String[kti.getTopics().size()]);

            String clientId = String.format("cli_%s_%s", config.getGroupId(), selectedSubscriberConfig.getUid());

            int receiveBufferConfig = 5 * 1024;
            if ("others".equalsIgnoreCase(this.configFile)) {
                receiveBufferConfig = 512;
            }

            logger.error("init with configFile={} receiveBufferConfig={};", this.configFile, receiveBufferConfig);

            KafkaSource<BinlogRawValue> binLogKafkaSource =
                    KafkaSource.<BinlogRawValue>builder()
                            .setBootstrapServers(kti.getKafkaConnection().getBootstrapServer())
                            .setStartingOffsets(OffsetsInitializer.committedOffsets(OffsetResetStrategy.EARLIEST))
                            .setGroupId(String.format("%s_%s", config.getGroupId(), selectedSubscriberConfig.getUid()))
                            .setClientIdPrefix(clientId)
                            .setTopics(selectedTopics)
                            .setDeserializer(kafkaRecordDeserializationSchema)
                            .setProperty(ConsumerConfig.SESSION_TIMEOUT_MS_CONFIG, "120000")
                            .setProperty(ConsumerConfig.MAX_POLL_INTERVAL_MS_CONFIG, "300000")
                            .setProperty(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, "true")
                            .setProperty(ConsumerConfig.CLIENT_ID_CONFIG, clientId)
                            .setProperty(ConsumerConfig.FETCH_MAX_BYTES_CONFIG, String.valueOf(20 * 1024 * 1024))
                            .setProperty(ConsumerConfig.RECEIVE_BUFFER_CONFIG, String.valueOf(receiveBufferConfig))
                            .setProperty("commit.offsets.on.checkpoint", "true")
                            .build();

            SingleOutputStreamOperator<BinlogRawValue> debeziumJsonNodeDtoKeyedStream
                    = env.fromSource(binLogKafkaSource, WatermarkStrategy.forMonotonousTimestamps(),
                    String.format("订阅-%s", selectedSubscriberConfig.getTopic())
            ).uid(selectedSubscriberConfig.getUid());

            buildJdbcSubscriber(
                    kti.getKafkaConnection().getBootstrapServer(),
                    debeziumJsonNodeDtoKeyedStream, jdbcExecutionOptions, config,
                    selectedSubscriberConfig.getSubscriberContents());
        });
    }
}
